@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-import errno
 import logging
 import socket
 from hashlib import md5
@@ -13,7 +12,7 @@ from botocore.exceptions import ClientError, ParamValidationError, BotoCoreError
 from s3fs.errors import translate_boto_error
 from s3fs.utils import ParamKwargsHelper
 
-logger = logging.getLogger(__name__)
+logger = logging.getLogger('s3fs')
 
 logging.getLogger('boto3').setLevel(logging.WARNING)
 logging.getLogger('botocore').setLevel(logging.WARNING)
@@ -100,12 +99,16 @@ class S3FileSystem(AbstractFileSystem):
     client_kwargs : dict of parameters for the boto3 client
     requester_pays : bool (False)
         If RequesterPays buckets are supported.
-    default_block_size: None, int
+    default_block_size: int (None)
         If given, the default block size value used for ``open()``, if no
         specific value is given at all time. The built-in default is 5MB.
     default_fill_cache : Bool (True)
         Whether to use cache filling with open by default. Refer to
         ``S3File.open``.
+    default_cache_type : string ('bytes')
+        If given, the default cache_type value used for ``open()``. Set to "none"
+        if no caching is desired. See fsspec's documentation for other available
+        cache_type values. Default cache_type is 'bytes'.
     version_aware : bool (False)
         Whether to support bucket versioning.  If enable this will require the
         user to have the necessary IAM permissions for dealing with versioned
@@ -135,7 +138,7 @@ class S3FileSystem(AbstractFileSystem):
     def __init__(self, anon=False, key=None, secret=None, token=None,
                  use_ssl=True, client_kwargs=None, requester_pays=False,
                  default_block_size=None, default_fill_cache=True,
-                 version_aware=False, config_kwargs=None,
+                 default_cache_type='bytes', version_aware=False, config_kwargs=None,
                  s3_additional_kwargs=None, session=None, username=None,
                  password=None, **kwargs):
         if key and username:
@@ -164,6 +167,7 @@ class S3FileSystem(AbstractFileSystem):
             config_kwargs = {}
         self.default_block_size = default_block_size or self.default_block_size
         self.default_fill_cache = default_fill_cache
+        self.default_cache_type = default_cache_type
         self.version_aware = version_aware
         self.client_kwargs = client_kwargs
         self.config_kwargs = config_kwargs
@@ -177,6 +181,9 @@ class S3FileSystem(AbstractFileSystem):
         return self._kwargs_helper.filter_dict(s3_method.__name__, kwargs)
 
     def _call_s3(self, method, *akwarglist, **kwargs):
+        kw2 = kwargs.copy()
+        kw2.pop('Body', None)
+        logger.debug("CALL: %s - %s - %s" % (method, akwarglist, kw2))
         additional_kwargs = self._get_s3_method_kwargs(method, *akwarglist,
                                                        **kwargs)
         return method(**additional_kwargs)
@@ -253,7 +260,7 @@ class S3FileSystem(AbstractFileSystem):
                 'token': cred['SessionToken'], 'anon': False}
 
     def _open(self, path, mode='rb', block_size=None, acl='', version_id=None,
-              fill_cache=None, cache_type='bytes', autocommit=True, **kwargs):
+              fill_cache=None, cache_type=None, autocommit=True, **kwargs):
         """ Open a file for reading or writing
 
         Parameters
@@ -280,7 +287,8 @@ class S3FileSystem(AbstractFileSystem):
             The encoding to use if opening the file in text mode. The platform's
             default text encoding is used if not given.
         cache_type : str
-            "bytes", "mmap" or "none"
+            See fsspec's documentation for available cache_type values. Set to "none"
+            if no caching is desired. If None, defaults to ``self.default_cache_type``.
         kwargs: dict-like
             Additional parameters used for s3 methods.  Typically used for
             ServerSideEncryption.
@@ -297,6 +305,9 @@ class S3FileSystem(AbstractFileSystem):
             raise ValueError("version_id cannot be specified if the filesystem "
                              "is not version aware")
 
+        if cache_type is None:
+            cache_type = self.default_cache_type
+
         return S3File(self, path, mode, block_size=block_size, acl=acl,
                       version_id=version_id, fill_cache=fill_cache,
                       s3_additional_kwargs=kw, cache_type=cache_type,
@@ -443,6 +454,14 @@ class S3FileSystem(AbstractFileSystem):
             except FileNotFoundError:
                 return False
 
+    def touch(self, path, truncate=True, **kwargs):
+        """Create empty file or truncate"""
+        bucket, key = split_path(path)
+        if not truncate and self.exists(path):
+            raise ValueError("S3 does not support touching existent files")
+        self._call_s3(self.s3.put_object, kwargs, Bucket=bucket, Key=key)
+        self.invalidate_cache(self._parent(path))
+
     def info(self, path, version_id=None):
         if path in ['/', '']:
             return {'name': path, 'size': 0, 'type': 'directory'}
@@ -929,8 +948,8 @@ class S3File(AbstractBufferedFile):
                 self.size = self.details['size']
             elif self.fs.version_aware:
                 self.version_id = self.details.get('VersionId')
-                # In this case we have not managed to get the VersionId out of details and 
-                # we should invalidate the cache and perform a full head_object since it 
+                # In this case we have not managed to get the VersionId out of details and
+                # we should invalidate the cache and perform a full head_object since it
                 # has likely been partially populated by ls.
                 if self.version_id is None:
                     self.fs.invalidate_cache(self.path)
@@ -952,6 +971,7 @@ class S3File(AbstractBufferedFile):
                                 **kwargs)
 
     def _initiate_upload(self):
+        logger.debug("Initiate upload for %s" % self)
         if self.acl and self.acl not in key_acls:
             raise ValueError('ACL not in %s', key_acls)
         self.parts = []
@@ -1041,6 +1061,7 @@ class S3File(AbstractBufferedFile):
                     (data0, data1) = (remainder[:partition], remainder[partition:])
 
             part = len(self.parts) + 1
+            logger.debug("Upload chunk %s, %s" % (self, part))
 
             for attempt in range(self.retries + 1):
                 try:
@@ -1065,7 +1086,12 @@ class S3File(AbstractBufferedFile):
             self.commit()
 
     def commit(self):
-        logger.debug("COMMIT")
+        logger.debug("Commit %s" % self)
+        if not self.parts:
+            logger.debug("Empty file committed %s" % self)
+            self._abort_mpu()
+            self.fs.touch(self.path)
+            return
         part_info = {'Parts': self.parts}
         write_result = self._call_s3(
             self.fs.s3.complete_multipart_upload,
@@ -1090,6 +1116,9 @@ class S3File(AbstractBufferedFile):
     def discard(self):
         if self.autocommit:
             raise ValueError("Cannot discard when autocommit is enabled")
+        self._abort_mpu()
+
+    def _abort_mpu(self):
         self._call_s3(
             self.fs.s3.abort_multipart_upload,
             Bucket=self.bucket,
@@ -1102,6 +1131,18 @@ def _fetch_range(client, bucket, key, version_id, start, end, max_attempts=10,
                  req_kw=None):
     if req_kw is None:
         req_kw = {}
+    if start == end:
+        # When these match, we would make a request with `range=start-end - 1`
+        # According to RFC2616, servers are supposed to ignore the Range
+        # field when it's invalid like this. S3 does ignore it, moto doesn't.
+        # To avoid differences in behavior under mocking, we just avoid
+        # making these requests. It's hoped that since we're being called
+        # from a caching object, this won't end up mattering.
+        logger.debug(
+            'skip fetch for negative range - bucket=%s,key=%s,start=%d,end=%d',
+            bucket, key, start, end
+        )
+        return b''
     logger.debug("Fetch: %s/%s, %s-%s", bucket, key, start, end)
     for i in range(max_attempts):
         try:
